# Copyright (c) 2013 - 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Unit Tests for Huawei FusionStorage drivers.
"""

import mock

from cinder import test
from cinder import utils
from cinder.volume.drivers.fusionstorage import fspythonapi


class FSPythonApiTestCase(test.TestCase):

    def setUp(self):
        super(FSPythonApiTestCase, self).setUp()
        self.api = fspythonapi.FSPythonApi()

    @mock.patch.object(fspythonapi.FSPythonApi, 'get_ip_port')
    @mock.patch.object(fspythonapi.FSPythonApi, 'get_manage_ip')
    @mock.patch.object(utils, 'execute')
    def test_start_execute_cmd(self, mock_execute,
                               mock_get_manage_ip, mock_get_ip_port):
        result1 = ['result=0\ndesc=success\n', '']
        result2 = ['result=50150007\ndesc=volume does not exist\n', '']
        result3 = ['result=50150008\ndesc=volume is being deleted\n', '']
        result4 = ['result=50\ndesc=exception\n', '']
        cmd = 'abcdef'

        mock_get_ip_port.return_value = ['127.0.0.1', '128.0.0.1']
        mock_get_manage_ip.return_value = '127.0.0.1'

        mock_execute.return_value = result1
        retval = self.api.start_execute_cmd(cmd, 0)
        self.assertEqual('result=0', retval)

        mock_execute.return_value = result2
        retval = self.api.start_execute_cmd(cmd, 0)
        self.assertEqual('result=0', retval)

        mock_execute.return_value = result3
        retval = self.api.start_execute_cmd(cmd, 0)
        self.assertEqual('result=0', retval)

        mock_execute.return_value = result4
        retval = self.api.start_execute_cmd(cmd, 0)
        self.assertEqual('result=50', retval)

        mock_execute.return_value = result1
        retval = self.api.start_execute_cmd(cmd, 1)
        self.assertEqual(['result=0', 'desc=success', ''], retval)

        mock_execute.return_value = result2
        retval = self.api.start_execute_cmd(cmd, 1)
        self.assertEqual('result=0', retval)

        mock_execute.return_value = result3
        retval = self.api.start_execute_cmd(cmd, 1)
        self.assertEqual('result=0', retval)

        mock_execute.return_value = result4
        retval = self.api.start_execute_cmd(cmd, 1)
        self.assertEqual(['result=50', 'desc=exception', ''], retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_create_volume(self, mock_start_execute):
        mock_start_execute.side_effect = ['result=0\n',
                                          'result=50150007\n', None]

        retval = self.api.create_volume('volume_name', 'pool_id-123', 1024, 0)
        self.assertEqual(0, retval)

        retval = self.api.create_volume('volume_name', 'pool_id-123', 1024, 0)
        self.assertEqual('50150007\n', retval)

        retval = self.api.create_volume('volume_name', 'pool_id-123', 1024, 0)
        self.assertEqual(1, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_extend_volume(self, mock_start_execute):
        mock_start_execute.side_effect = ['result=0\n',
                                          'result=50150007\n', None]

        retval = self.api.extend_volume('volume_name', 1024)
        self.assertEqual(0, retval)

        retval = self.api.extend_volume('volume_name', 1024)
        self.assertEqual('50150007\n', retval)

        retval = self.api.extend_volume('volume_name', 1024)
        self.assertEqual(1, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_create_volume_from_snap(self, mock_start_execute):
        mock_start_execute.side_effect = ['result=0\n',
                                          'result=50150007\n', None]

        retval = self.api.create_volume_from_snap('volume_name', 1024,
                                                  'snap_name')
        self.assertEqual(0, retval)

        retval = self.api.create_volume_from_snap('volume_name', 1024,
                                                  'snap_name')
        self.assertEqual('50150007\n', retval)

        retval = self.api.create_volume_from_snap('volume_name', 1024,
                                                  'snap_name')
        self.assertEqual(1, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_create_fullvol_from_snap(self, mock_start_execute):
        mock_start_execute.side_effect = ['result=0\n',
                                          'result=50150007\n', None]

        retval = self.api.create_fullvol_from_snap('volume_name', 'snap_name')
        self.assertEqual(0, retval)

        retval = self.api.create_fullvol_from_snap('volume_name', 'snap_name')
        self.assertEqual('50150007\n', retval)

        retval = self.api.create_fullvol_from_snap('volume_name', 'snap_name')
        self.assertEqual(1, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'create_snapshot')
    @mock.patch.object(fspythonapi.FSPythonApi, 'create_volume')
    @mock.patch.object(fspythonapi.FSPythonApi, 'delete_snapshot')
    @mock.patch.object(fspythonapi.FSPythonApi, 'delete_volume')
    @mock.patch.object(fspythonapi.FSPythonApi, 'create_fullvol_from_snap')
    def test_create_volume_from_volume(self, mock_create_fullvol,
                                       mock_delete_volume, mock_delete_snap,
                                       mock_create_volume, mock_create_snap):
        mock_create_snap.return_value = 0
        mock_create_volume.return_value = 0
        mock_create_fullvol.return_value = 0

        retval = self.api.create_volume_from_volume('vol_name', 1024,
                                                    'src_vol_name')
        self.assertEqual(0, retval)

        mock_create_snap.return_value = 1
        retval = self.api.create_volume_from_volume('vol_name', 1024,
                                                    'src_vol_name')
        self.assertEqual(1, retval)

        mock_create_snap.return_value = 0
        mock_create_volume.return_value = 1
        retval = self.api.create_volume_from_volume('vol_name', 1024,
                                                    'src_vol_name')
        self.assertEqual(1, retval)

        mock_create_volume.return_value = 0
        self.api.create_fullvol_from_snap.return_value = 1
        retval = self.api.create_volume_from_volume('vol_name', 1024,
                                                    'src_vol_name')
        self.assertEqual(1, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'create_snapshot')
    @mock.patch.object(fspythonapi.FSPythonApi, 'create_volume_from_snap')
    def test_create_clone_volume_from_volume(self, mock_volume, mock_snap):
        mock_snap.side_effect = [0, 1]
        mock_volume.side_effect = [0, 1]
        retval = self.api.create_clone_volume_from_volume('vol_name', 1024,
                                                          'src_vol_name')
        self.assertEqual(0, retval)
        retval = self.api.create_clone_volume_from_volume('vol_name', 1024,
                                                          'src_vol_name')
        self.assertEqual(1, retval)

    def test_volume_info_analyze_success(self):
        vol_info = ('vol_name=vol1,father_name=vol1_father,'
                    'status=available,vol_size=1024,real_size=1024,'
                    'pool_id=pool1,create_time=01/01/2015')
        vol_info_res = {'result': 0, 'vol_name': 'vol1',
                        'father_name': 'vol1_father',
                        'status': 'available', 'vol_size': '1024',
                        'real_size': '1024', 'pool_id': 'pool1',
                        'create_time': '01/01/2015'}

        retval = self.api.volume_info_analyze(vol_info)
        self.assertEqual(vol_info_res, retval)

    def test_volume_info_analyze_fail(self):
        vol_info = ''
        vol_info_res = {'result': 1, 'vol_name': '', 'father_name': '',
                        'status': '', 'vol_size': '', 'real_size': '',
                        'pool_id': '', 'create_time': ''}
        retval = self.api.volume_info_analyze(vol_info)
        self.assertEqual(vol_info_res, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    @mock.patch.object(fspythonapi.FSPythonApi, 'volume_info_analyze')
    @mock.patch.object(fspythonapi.FSPythonApi, 'delete_snapshot')
    def test_query_volume(self, mock_delete, mock_analyze, mock_execute):
        exec_result = ['result=0\n',
                       'vol_name=vol1,father_name=vol1_father,status=0,' +
                       'vol_size=1024,real_size=1024,pool_id=pool1,' +
                       'create_time=01/01/2015']
        query_result = {'result': 0, 'vol_name': 'vol1',
                        'father_name': 'vol1_father', 'status': '0',
                        'vol_size': '1024', 'real_size': '1024',
                        'pool_id': 'pool1', 'create_time': '01/01/2015'}
        mock_delete.return_value = 0
        mock_execute.return_value = exec_result
        mock_analyze.return_value = query_result
        retval = self.api.query_volume('vol1')
        self.assertEqual(query_result, retval)

        exec_result = ['result=0\n',
                       'vol_name=vol1,father_name=vol1_father,status=1,' +
                       'vol_size=1024,real_size=1024,pool_id=pool1,' +
                       'create_time=01/01/2015']
        query_result = {'result': 0, 'vol_name': 'vol1',
                        'father_name': 'vol1_father', 'status': '1',
                        'vol_size': '1024', 'real_size': '1024',
                        'pool_id': 'pool1', 'create_time': '01/01/2015'}
        mock_delete.return_value = 0
        mock_execute.return_value = exec_result
        mock_analyze.return_value = query_result
        retval = self.api.query_volume('vol1')
        self.assertEqual(query_result, retval)

        vol_info_failure = 'result=32500000\n'
        failure_res = {'result': 1, 'vol_name': '', 'father_name': '',
                       'status': '', 'vol_size': '', 'real_size': '',
                       'pool_id': '', 'create_time': ''}
        mock_execute.return_value = vol_info_failure
        retval = self.api.query_volume('vol1')
        self.assertEqual(failure_res, retval)

        vol_info_failure = None
        failure_res = {'result': 1, 'vol_name': '', 'father_name': '',
                       'status': '', 'vol_size': '', 'real_size': '',
                       'pool_id': '', 'create_time': ''}

        mock_execute.return_value = vol_info_failure
        retval = self.api.query_volume('vol1')
        self.assertEqual(failure_res, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_delete_volume(self, mock_execute):
        mock_execute.side_effect = ['result=0\n',
                                    'result=50150007\n', None]

        retval = self.api.delete_volume('volume_name')
        self.assertEqual(0, retval)

        retval = self.api.delete_volume('volume_name')
        self.assertEqual('50150007\n', retval)

        retval = self.api.delete_volume('volume_name')
        self.assertEqual(1, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_create_snapshot(self, mock_execute):
        mock_execute.side_effect = ['result=0\n',
                                    'result=50150007\n', None]

        retval = self.api.create_snapshot('snap_name', 'vol_name', 0)
        self.assertEqual(0, retval)

        retval = self.api.create_snapshot('snap_name', 'vol_name', 0)
        self.assertEqual('50150007\n', retval)

        retval = self.api.create_snapshot('snap_name', 'vol_name', 0)
        self.assertEqual(1, retval)

    def test_snap_info_analyze_success(self):
        snap_info = ('snap_name=snap1,father_name=snap1_father,status=0,'
                     'snap_size=1024,real_size=1024,pool_id=pool1,'
                     'delete_priority=1,create_time=01/01/2015')
        snap_info_res = {'result': 0, 'snap_name': 'snap1',
                         'father_name': 'snap1_father', 'status': '0',
                         'snap_size': '1024', 'real_size': '1024',
                         'pool_id': 'pool1', 'delete_priority': '1',
                         'create_time': '01/01/2015'}

        retval = self.api.snap_info_analyze(snap_info)
        self.assertEqual(snap_info_res, retval)

    def test_snap_info_analyze_fail(self):
        snap_info = ''
        snap_info_res = {'result': 1, 'snap_name': '', 'father_name': '',
                         'status': '', 'snap_size': '', 'real_size': '',
                         'pool_id': '', 'delete_priority': '',
                         'create_time': ''}
        retval = self.api.snap_info_analyze(snap_info)
        self.assertEqual(snap_info_res, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_query_snap(self, mock_execute):
        exec_result = ['result=0\n',
                       'snap_name=snap1,father_name=snap1_father,status=0,' +
                       'snap_size=1024,real_size=1024,pool_id=pool1,' +
                       'delete_priority=1,create_time=01/01/2015']
        query_result = {'result': 0, 'snap_name': 'snap1',
                        'father_name': 'snap1_father', 'status': '0',
                        'snap_size': '1024', 'real_size': '1024',
                        'pool_id': 'pool1', 'delete_priority': '1',
                        'create_time': '01/01/2015'}
        mock_execute.return_value = exec_result
        retval = self.api.query_snap('snap1')
        self.assertEqual(query_result, retval)

        exec_result = ['result=50150007\n']
        qurey_result = {'result': '50150007\n', 'snap_name': '',
                        'father_name': '', 'status': '', 'snap_size': '',
                        'real_size': '', 'pool_id': '',
                        'delete_priority': '', 'create_time': ''}
        mock_execute.return_value = exec_result
        retval = self.api.query_snap('snap1')
        self.assertEqual(qurey_result, retval)

        exec_result = ''
        query_result = {'result': 1, 'snap_name': '', 'father_name': '',
                        'status': '', 'snap_size': '', 'real_size': '',
                        'pool_id': '', 'delete_priority': '',
                        'create_time': ''}
        mock_execute.return_value = exec_result
        retval = self.api.query_snap('snap1')
        self.assertEqual(query_result, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_delete_snapshot(self, mock_execute):
        mock_execute.side_effect = ['result=0\n',
                                    'result=50150007\n', None]

        retval = self.api.delete_snapshot('snap_name')
        self.assertEqual(0, retval)

        retval = self.api.delete_snapshot('snap_name')
        self.assertEqual('50150007\n', retval)

        retval = self.api.delete_snapshot('snap_name')
        self.assertEqual(1, retval)

    def test_pool_info_analyze(self):
        pool_info = 'pool_id=pool100,total_capacity=1024,' + \
                    'used_capacity=500,alloc_capacity=500'
        analyze_res = {'result': 0, 'pool_id': 'pool100',
                       'total_capacity': '1024', 'used_capacity': '500',
                       'alloc_capacity': '500'}

        retval = self.api.pool_info_analyze(pool_info)
        self.assertEqual(analyze_res, retval)

        pool_info = ''
        analyze_res = {'result': 1, 'pool_id': '', 'total_capacity': '',
                       'used_capacity': '', 'alloc_capacity': ''}
        retval = self.api.pool_info_analyze(pool_info)
        self.assertEqual(analyze_res, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_query_pool_info(self, mock_execute):
        exec_result = ['result=0\n',
                       'pool_id=0,total_capacity=1024,' +
                       'used_capacity=500,alloc_capacity=500\n']
        query_result = {'result': 0, 'pool_id': '0',
                        'total_capacity': '1024', 'used_capacity': '500',
                        'alloc_capacity': '500'}
        mock_execute.return_value = exec_result
        retval = self.api.query_pool_info('0')
        self.assertEqual(query_result, retval)

        exec_result = ['result=51050008\n']
        query_result = {'result': '51050008\n', 'pool_id': '',
                        'total_capacity': '', 'used_capacity': '',
                        'alloc_capacity': ''}
        mock_execute.return_value = exec_result
        retval = self.api.query_pool_info('0')
        self.assertEqual(query_result, retval)

        exec_result = ''
        query_result = {'result': 1, 'pool_id': '', 'total_capacity': '',
                        'used_capacity': '', 'alloc_capacity': ''}
        mock_execute.return_value = exec_result
        retval = self.api.query_pool_info('0')
        self.assertEqual(query_result, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_query_pool_type(self, mock_execute):
        exec_result = ['result=0\n',
                       'pool_id=0,total_capacity=1024,' +
                       'used_capacity=500,alloc_capacity=500\n']
        query_result = (0, [{'result': 0,
                             'pool_id': '0', 'total_capacity': '1024',
                             'used_capacity': '500', 'alloc_capacity': '500'}])

        mock_execute.return_value = exec_result
        retval = self.api.query_pool_type('sata2copy')
        self.assertEqual(query_result, retval)

        exec_result = ['result=0\n',
                       'pool_id=0,total_capacity=1024,' +
                       'used_capacity=500,alloc_capacity=500\n',
                       'pool_id=1,total_capacity=2048,' +
                       'used_capacity=500,alloc_capacity=500\n']
        query_result = (0, [{'result': 0, 'pool_id': '0',
                             'total_capacity': '1024', 'used_capacity': '500',
                             'alloc_capacity': '500'},
                            {'result': 0, 'pool_id': '1',
                             'total_capacity': '2048', 'used_capacity': '500',
                             'alloc_capacity': '500'}])
        mock_execute.return_value = exec_result
        retval = self.api.query_pool_type('sata2copy')
        self.assertEqual(query_result, retval)

        exec_result = ['result=51010015\n']
        query_result = (51010015, [])
        mock_execute.return_value = exec_result
        retval = self.api.query_pool_type('sata2copy')
        self.assertEqual(query_result, retval)

        exec_result = ''
        query_result = (0, [])
        mock_execute.return_value = exec_result
        retval = self.api.query_pool_type('sata2copy')
        self.assertEqual(query_result, retval)

    @mock.patch.object(fspythonapi.FSPythonApi, 'start_execute_cmd')
    def test_query_dsware_version(self, mock_execute):
        mock_execute.side_effect = ['result=0\n', 'result=50500001\n',
                                    'result=50150007\n', None]

        retval = self.api.query_dsware_version()
        self.assertEqual(0, retval)

        retval = self.api.query_dsware_version()
        self.assertEqual(1, retval)

        retval = self.api.query_dsware_version()
        self.assertEqual('50150007\n', retval)

        retval = self.api.query_dsware_version()
        self.assertEqual(2, retval)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Tests for Volume retype Code."""

import mock
from oslo_config import cfg

from cinder import context
from cinder import exception
from cinder import objects
from cinder.policies import volume_actions as vol_action_policies
from cinder.policies import volumes as volume_policies
from cinder import quota
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import utils as tests_utils
from cinder.tests.unit import volume as base
from cinder.volume import volume_types


QUOTAS = quota.QUOTAS

CONF = cfg.CONF


class VolumeRetypeTestCase(base.BaseVolumeTestCase):

    def setUp(self):
        super(VolumeRetypeTestCase, self).setUp()
        self.patch('cinder.volume.utils.clear_volume', autospec=True)
        self.expected_status = 'available'
        self.service_id = 1
        self.user_context = context.RequestContext(user_id=fake.USER_ID,
                                                   project_id=fake.PROJECT_ID)

        volume_types.create(self.context,
                            "fake_vol_type",
                            {},
                            description="fake_type")
        volume_types.create(self.context,
                            "multiattach-type",
                            {'multiattach': "<is> True"},
                            description="test-multiattach")
        volume_types.create(self.context,
                            "multiattach-type2",
                            {'multiattach': "<is> True"},
                            description="test-multiattach")
        self.default_vol_type = objects.VolumeType.get_by_name_or_id(
            self.context,
            'fake_vol_type')
        self.multiattach_type = objects.VolumeType.get_by_name_or_id(
            self.context,
            'multiattach-type')
        self.multiattach_type2 = objects.VolumeType.get_by_name_or_id(
            self.context,
            'multiattach-type2')

    def fake_get_vtype(self, context, identifier):
        if identifier == "multiattach-type":
            return self.multiattach_type
        elif identifier == 'multiattach-type2':
            return self.multiattach_type2
        else:
            return self.default_vol_type

    @mock.patch('cinder.context.RequestContext.authorize')
    @mock.patch.object(volume_types, 'get_by_name_or_id')
    def test_retype_multiattach(self, _mock_get_types, mock_authorize):
        """Verify multiattach retype restrictions."""

        _mock_get_types.side_effect = self.fake_get_vtype

        # Test going from default type to multiattach
        vol = self.volume_api.create(self.context,
                                     1,
                                     'test-vol',
                                     '')

        vol.update({'status': 'available'})
        vol.save()
        self.volume_api.retype(self.user_context,
                               vol,
                               'multiattach-type')
        vol = objects.Volume.get_by_id(self.context, vol.id)
        self.assertTrue(vol.multiattach)

        # Test going from multiattach to a non-multiattach type
        vol = self.volume_api.create(
            self.context,
            1,
            'test-multiattachvol',
            '',
            volume_type=self.multiattach_type)
        vol.update({'status': 'available'})
        vol.save()
        self.volume_api.retype(self.user_context,
                               vol,
                               'fake_vol_type')
        vol = objects.Volume.get_by_id(self.context, vol.id)
        self.assertFalse(vol.multiattach)

        # Test trying to retype an in-use volume
        vol.update({'status': 'in-use'})
        vol.save()
        self.assertRaises(exception.InvalidInput,
                          self.volume_api.retype,
                          self.context,
                          vol,
                          'multiattach-type')
        mock_authorize.assert_has_calls(
            [mock.call(volume_policies.CREATE_POLICY),
             mock.call(volume_policies.CREATE_POLICY),
             mock.call(vol_action_policies.RETYPE_POLICY, target_obj=mock.ANY),
             mock.call(volume_policies.MULTIATTACH_POLICY,
                       target_obj=mock.ANY),
             mock.call(volume_policies.CREATE_POLICY),
             mock.call(volume_policies.CREATE_POLICY),
             mock.call(vol_action_policies.RETYPE_POLICY, target_obj=mock.ANY),
             mock.call(vol_action_policies.RETYPE_POLICY, target_obj=mock.ANY),
             ])

    @mock.patch('cinder.context.RequestContext.authorize')
    def test_multiattach_to_multiattach_retype(self, mock_authorize):
        # Test going from multiattach to multiattach

        vol = tests_utils.create_volume(self.context,
                                        multiattach=True,
                                        volume_type_id=
                                        self.multiattach_type.id)

        self.assertTrue(vol.multiattach)
        self.volume_api.retype(self.user_context,
                               vol,
                               'multiattach-type2')
        vol.refresh()
        self.assertTrue(vol.multiattach)

        mock_authorize.assert_has_calls(
            [mock.call(vol_action_policies.RETYPE_POLICY, target_obj=mock.ANY)
             ])



# Copyright 2016 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import requests
import time

from oslo_log import log as logging
from oslo_serialization import jsonutils

from cinder import exception
from cinder.i18n import _
from cinder.utils import retry

LOG = logging.getLogger(__name__)
TIMEOUT = 60


def check_error(response):
    code = response.status_code
    if code not in (200, 201, 202):
        reason = response.reason
        body = response.content
        try:
            content = jsonutils.loads(body) if body else None
        except ValueError:
            raise exception.VolumeBackendAPIException(
                data=_(
                    'Could not parse response: %(code)s %(reason)s '
                    '%(content)s') % {
                        'code': code, 'reason': reason, 'content': body})
        if content and 'code' in content:
            raise exception.NexentaException(content)
        raise exception.VolumeBackendAPIException(
            data=_(
                'Got bad response: %(code)s %(reason)s %(content)s') % {
                    'code': code, 'reason': reason, 'content': content})


class RESTCaller(object):

    retry_exc_tuple = (
        requests.exceptions.ConnectionError,
        requests.exceptions.ConnectTimeout
    )

    def __init__(self, proxy, method):
        self.__proxy = proxy
        self.__method = method

    def get_full_url(self, path):
        return '/'.join((self.__proxy.url, path))

    @retry(retry_exc_tuple, interval=1, retries=6)
    def __call__(self, *args):
        url = self.get_full_url(args[0])
        kwargs = {'timeout': TIMEOUT, 'verify': False}
        data = None
        if len(args) > 1:
            data = args[1]
            kwargs['json'] = data

        LOG.debug('Sending JSON data: %s, method: %s, data: %s',
                  url, self.__method, data)

        response = getattr(self.__proxy.session, self.__method)(url, **kwargs)
        check_error(response)
        content = (jsonutils.loads(response.content)
                   if response.content else None)
        LOG.debug("Got response: %(code)s %(reason)s %(content)s", {
            'code': response.status_code,
            'reason': response.reason,
            'content': content})

        if response.status_code == 202 and content:
            url = self.get_full_url(content['links'][0]['href'])
            keep_going = True
            while keep_going:
                time.sleep(1)
                response = self.__proxy.session.get(url, verify=False)
                check_error(response)
                LOG.debug("Got response: %(code)s %(reason)s", {
                    'code': response.status_code,
                    'reason': response.reason})
                content = response.json() if response.content else None
                keep_going = response.status_code == 202
        return content


class HTTPSAuth(requests.auth.AuthBase):

    def __init__(self, url, username, password):
        self.url = url
        self.username = username
        self.password = password
        self.token = None

    def __eq__(self, other):
        return all([
            self.url == getattr(other, 'url', None),
            self.username == getattr(other, 'username', None),
            self.password == getattr(other, 'password', None),
            self.token == getattr(other, 'token', None)
        ])

    def __ne__(self, other):
        return not self == other

    def handle_401(self, r, **kwargs):
        if r.status_code == 401:
            LOG.debug('Got 401. Trying to reauth...')
            self.token = self.https_auth()
            # Consume content and release the original connection
            # to allow our new request to reuse the same one.
            r.content
            r.close()
            prep = r.request.copy()
            requests.cookies.extract_cookies_to_jar(
                prep._cookies, r.request, r.raw)
            prep.prepare_cookies(prep._cookies)

            prep.headers['Authorization'] = 'Bearer %s' % self.token
            _r = r.connection.send(prep, **kwargs)
            _r.history.append(r)
            _r.request = prep

            return _r
        return r

    def __call__(self, r):
        if not self.token:
            self.token = self.https_auth()
        r.headers['Authorization'] = 'Bearer %s' % self.token
        r.register_hook('response', self.handle_401)
        return r

    def https_auth(self):
        LOG.debug('Sending auth request...')
        url = '/'.join((self.url, 'auth/login'))
        headers = {'Content-Type': 'application/json'}
        data = {'username': self.username, 'password': self.password}
        response = requests.post(url, json=data, verify=False,
                                 headers=headers, timeout=TIMEOUT)
        check_error(response)
        response.close()
        if response.content:
            content = jsonutils.loads(response.content)
            token = content['token']
            del content['token']
            LOG.debug("Got response: %(code)s %(reason)s %(content)s", {
                'code': response.status_code,
                'reason': response.reason,
                'content': content})
            return token
        raise exception.VolumeBackendAPIException(
            data=_(
                'Got bad response: %(code)s %(reason)s') % {
                    'code': response.status_code, 'reason': response.reason})


class NexentaJSONProxy(object):

    def __init__(self, host, port, user, password, use_https):
        self.session = requests.Session()
        self.session.headers.update({'Content-Type': 'application/json'})
        self.host = host
        if use_https:
            self.scheme = 'https'
            self.port = port if port else 8443
            self.session.auth = HTTPSAuth(self.url, user, password)
        else:
            self.scheme = 'http'
            self.port = port if port else 8080
            self.session.auth = (user, password)

    @property
    def url(self):
        return '%(scheme)s://%(host)s:%(port)s' % {
            'scheme': self.scheme,
            'host': self.host,
            'port': self.port}

    def __getattr__(self, name):
        if name in ('get', 'post', 'put', 'delete'):
            return RESTCaller(self, name)
        return super(NexentaJSONProxy, self).__getattribute__(name)

    def __repr__(self):
        return 'HTTP JSON proxy: %s' % self.url



# Copyright (c) 2017 Veritas Technologies LLC.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from cinder import context
from cinder import exception
from cinder import test
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.volume import configuration as conf
from cinder.volume.drivers.veritas import vrtshyperscale as vrts


class FakeDb(object):
    def volume_metadata_get(self, *a, **kw):
        return {}

    def volume_metadata_update(self, *a, **kw):
        return None


def _stub_volume(*args, **kwargs):
    updates = {'provider_location': 'hyperscale-sv:/hyperscale'}
    return fake_volume.fake_db_volume(**updates)


def _stub_snapshot(*args, **kwargs):
    updates = {'volume': _stub_volume(), 'name': 'vrts'}
    return fake_snapshot.fake_db_snapshot(**updates)


def _stub_stats():
    data = {}
    data["volume_backend_name"] = 'Veritas_HyperScale'
    data["vendor_name"] = 'Veritas Technologies LLC'
    data["driver_version"] = '1.0'
    data["storage_protocol"] = 'nfs'
    data['total_capacity_gb'] = 0.0
    data['free_capacity_gb'] = 0.0
    data['reserved_percentage'] = 0
    data['QoS_support'] = False
    return data


class VRTSHyperScaleDriverTestCase(test.TestCase):
    """Test case for Veritas HyperScale VolumeDriver."""

    driver_name = "cinder.volume.drivers.veritas.vrtshyperscale"

    @staticmethod
    def gvmv_side_effect(arg1, arg2):
        """Mock side effect for _get_volume_metadata_value."""
        # mock the return of get_volume_metadata_value
        # for different arguments
        if arg2 == 'Secondary_datanode_key':
            return '{9876}'
        elif arg2 == 'Secondary_datanode_ip':
            return '192.0.2.2'
        elif arg2 == 'current_dn_ip':
            return '192.0.2.1'
        elif arg2 == 'vsa_ip':
            return '192.0.2.1'

    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._fetch_config_for_compute')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._fetch_config_for_datanode')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._fetch_config_for_controller')
    def setUp(self, mock_fcfcntr, mock_fcfd, mock_fcfc):
        mock_fcfcntr.return_value = None
        mock_fcfd.return_value = None
        mock_fcfc.return_value = None

        # Initialise a test seup
        super(VRTSHyperScaleDriverTestCase, self).setUp()

        self.configuration = mock.Mock(conf.Configuration(None))
        self.configuration.reserved_percentage = 0
        self.context = context.get_admin_context()
        self.driver = vrts.HyperScaleDriver(
            db=FakeDb(), configuration=self.configuration)
        self.driver.dn_routing_key = '{1234}'
        self.driver.datanode_ip = '192.0.2.1'
        self.volume = _stub_volume()
        self.snapshot = _stub_snapshot()

    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_replicas')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_details_for_create_volume')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    @mock.patch('cinder.volume.api.API.update_volume_metadata')
    def test_create_volume_single_replicas(self, mock_uvm, mock_mdp,
                                           mock_gvdfcv, mock_get_replicas,
                                           mock_gvm):
        """Test single volume replica. Happy path test case."""
        # Mock volume meatadata
        mock_gvm.return_value = _stub_volume()

        # Mock number of replicas to 1
        mock_get_replicas.return_value = 1
        # assume volume details are populated correctly
        mock_gvdfcv.return_value = _stub_volume()

        # assume volume message is sent to data node successfully
        mock_mdp.return_value = ("", None)
        # assume that the volume metadata gets updated correctly
        mock_uvm.return_value = {}

        # declare the expected result
        expected_result = {
            'provider_location': 'hyperscale-sv:/hyperscale',
            'metadata': mock_gvm.return_value
        }

        # call create volume and get the result
        actual_result = self.driver.create_volume(self.volume)

        # Test if the return value matched the expected results
        self.assertDictEqual(actual_result, expected_result)

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.get_hyperscale_version')
    def test_check_for_setup_error(self, mock_ghv):
        """Test check for setup errors in Veritas HyperScale driver.

        The test case checks happy path execution when driver version 1.0.0
        is installed.
        """
        mock_ghv.return_value = "1.0.0"

        # check the driver for setup errors
        self.driver.check_for_setup_error()

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.get_hyperscale_version')
    def test_check_for_setup_error_unsupported_version(self, mock_ghv):
        """Test check for setup errors in Veritas HyperScale driver.

        The test case checks happy path execution when driver version 1.0.0
        is installed.
        """
        mock_ghv.return_value = "1.0.0.1"

        # check the driver for setup errors
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.check_for_setup_error)

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.get_hyperscale_version')
    def test_check_for_setup_error_exception(self, mock_ghv):
        """Test check for setup errors in Veritas HyperScale driver.

        The test case checks happy path execution when driver version 1.0.0
        is installed.
        """
        mock_ghv.side_effect = exception.ErrorInHyperScaleVersion(
            cmd_error="mock error")

        # check the driver for setup errors
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.check_for_setup_error)

    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata_value')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    def test_delete_volume_no_replica(self, mock_mdp, mock_gvmv):
        """Test happy path for delete_volume one data nodes."""
        mock_gvmv.return_value = None
        self.driver.delete_volume(self.volume)

        message_body = {'display_name': self.volume['name']}

        mock_mdp.assert_called_with(self.driver.dn_routing_key,
                                    'hyperscale.storage.dm.volume.delete',
                                    **message_body)

    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata_value')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    def test_delete_volume_more_than_one_replica(self, mock_mdp, mock_gvmv):
        """Test happy path for delete_volume with more than one data nodes."""
        mock_gvmv.side_effect = VRTSHyperScaleDriverTestCase.gvmv_side_effect

        message_body = {'display_name': self.volume['name']}

        # make the delete call
        self.driver.delete_volume(self.volume)

        # check if delete volume sent to reflection target on data node
        # check if mq message sent with 'Secondary_datanode_key'
        mock_mdp.assert_any_call(
            '{9876}', 'hyperscale.storage.dm.volume.delete', **message_body)

        # check if the delete is sent to primary data node as well
        mock_mdp.assert_any_call(self.driver.dn_routing_key,
                                 'hyperscale.storage.dm.volume.delete',
                                 **message_body)

    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata_value')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    def test_delete_volume_no_replica_failure(self, mock_mdp, mock_gvmv):
        """Failure case for delete_volume one node in data plane."""
        mock_gvmv.side_effect = None
        self.driver.delete_volume(self.volume)
        mock_mdp.side_effect = exception.UnableToProcessHyperScaleCmdOutput(
            cmd_out='mock error')
        self.assertRaises(exception.VolumeIsBusy, self.driver.delete_volume,
                          self.volume)

    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata_value')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    def test_delete_volume_more_than_one_replica_failure(self, mock_mdp,
                                                         mock_gvmv):
        """failure case for delete_volume with more than one data nodes."""
        mock_gvmv.side_effect = VRTSHyperScaleDriverTestCase.gvmv_side_effect

        mock_mdp.side_effect = exception.UnableToProcessHyperScaleCmdOutput(
            cmd_out='mock error')

        self.assertRaises(exception.VolumeIsBusy, self.driver.delete_volume,
                          self.volume)

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.get_guid_with_curly_brackets')
    def test_delete_snapshot_force_flag(self, mock_ggwcb):
        """Test snapshot deletion does not happen if force flag is set."""
        # get a mock snapshot object
        snapshot = fake_snapshot.fake_db_snapshot()
        # set the force in metadata of snapshot
        snapshot['metadata'] = {"force": "force"}

        # call the delete volume
        self.driver.delete_snapshot(snapshot)

        # if snapshot has force set in metadata then
        # get_guid_with_curly_brackets() will not be called because we
        # return as soon as we see force
        mock_ggwcb.assert_not_called()

    def test_delete_snapshot_isbusy_flag(self):
        """Test snapshot deletion throws exception if snapshot is busy."""
        # get a mock snapshot object
        snapshot = fake_snapshot.fake_db_snapshot()
        # set the force in metadata of snapshot
        snapshot['metadata'] = {"is_busy": "is_busy"}

        # call the delete volume to check if it raises Busy Exception
        self.assertRaises(exception.SnapshotIsBusy,
                          self.driver.delete_snapshot, snapshot)

    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata')
    @mock.patch('cinder.volume.api.API.get_volume')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    def test_delete_snapshot_from_primary_dn(self, mock_mdp, mock_gv,
                                             mock_gvm):
        """Test snapshot deletion from primary DN."""
        # get mock volume
        mock_gv.return_value = None
        mock_gvm.return_value = {'current_dn_ip': self.driver.datanode_ip}

        message_body = {}
        message_body['volume_guid'] = '{' + self.volume['id'] + '}'
        message_body['snapshot_id'] = '{' + self.snapshot['id'] + '}'

        # call delete snapshot
        self.driver.delete_snapshot(self.snapshot)

        # assert msg is sent over mq with primary DN routing key
        mock_mdp.assert_called_with(self.driver.dn_routing_key,
                                    'hyperscale.storage.dm.version.delete',
                                    **message_body)

    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata')
    @mock.patch('cinder.volume.api.API.get_volume')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata_value')
    def test_delete_snapshot_from_current_dn(self, mock_gvmv, mock_mdp,
                                             mock_gv, mock_gvm):
        """Test snapshot deletion DN value from volume."""
        # get a mock volume
        mock_gv.return_value = _stub_volume()

        # get a mock value of DN from volume
        mock_gvmv.return_value = '{9876}'

        message_body = {}
        message_body['volume_guid'] = '{' + self.volume['id'] + '}'
        message_body['snapshot_id'] = '{' + self.snapshot['id'] + '}'

        # call delete snapshot
        self.driver.delete_snapshot(self.snapshot)

        # assert msg is sent over mq with key from volume's current_dn_owner
        mock_mdp.assert_called_with(
            '{9876}', 'hyperscale.storage.dm.version.delete', **message_body)

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    def test_fetch_volume_stats_failure(self, mock_mdp):
        """Test case checking failure of pool for fetching stats."""
        # since we have initialised the pool to None in setup()
        # the function will return only the stub without populating
        # any free and used stats
        mock_obj = {'payload': {}}

        mock_mdp.return_value = (mock_obj, None)
        self.assertDictEqual(_stub_stats(), self.driver._fetch_volume_status())

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    def test_create_cloned_volume_with_exception(self, mock_mdp):
        """Test case throws exception when command failed to execute."""
        vol_a = _stub_volume()
        vol_b = _stub_volume()
        mock_mdp.side_effect = exception.UnableToExecuteHyperScaleCmd(
            cmd_out='mock error')
        self.assertRaises(exception.UnableToExecuteHyperScaleCmd,
                          self.driver.create_cloned_volume, vol_b, vol_a)

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale'
                '.HyperScaleDriver._select_rt')
    def test_create_cloned_volume_with_no_replica(self, mock_srt, mock_mdp):
        """Test case clone volume when there is no replica."""
        mock_obj = {'payload': {}}
        mock_mdp.return_value = (mock_obj, None)
        mock_srt.return_value = (None, None)
        vol_a = _stub_volume()
        vol_b = _stub_volume()
        self.assertDictContainsSubset({
            'provider_location': 'hyperscale-sv:/hyperscale'
        }, self.driver.create_cloned_volume(vol_b, vol_a))

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale'
                '.HyperScaleDriver._select_rt')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata_value')
    def test_create_cloned_volume_with_replica(self, mock_gvmv, mock_srt,
                                               mock_mdp):
        """Test case clone volume when there is replica."""
        mock_gvmv.side_effect = VRTSHyperScaleDriverTestCase.gvmv_side_effect
        mock_obj = {'payload': {}}
        mock_mdp.return_value = (mock_obj, None)
        mock_srt.return_value = ('{1234}', '192.0.2.2')
        vol_a = _stub_volume()
        vol_b = _stub_volume()
        metadata = {
            'current_dn_owner': '{1234}',
            'Potential_secondary_key': '{1234}',
            'Primary_datanode_ip': '192.0.2.1',
            'Potential_secondary_ip': '192.0.2.2',
            'current_dn_ip': '192.0.2.1',
            'source_volid': vol_a['id'],
            'size': vol_a['size']
        }
        self.assertDictContainsSubset({
            'provider_location': 'hyperscale-sv:/hyperscale',
            'metadata': metadata
        }, self.driver.create_cloned_volume(vol_b, vol_a))

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    def test_extend_volume_with_exception(self, mock_mdp):
        """Test case extend volume to the given size in GB."""
        mock_mdp.side_effect = exception.UnableToProcessHyperScaleCmdOutput(
            cmd_out='mock error')
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.extend_volume, _stub_volume(), 256)

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    def test_extend_volume_no_exception(self, mock_mdp):
        """Test case extend volume thorws exception."""
        mock_mdp.return_value = (None, None)
        self.driver.extend_volume(_stub_volume(), 256)

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    def test_create_volume_from_snapshot_with_exception(self, mock_mdp):
        """Test case create volume from snapshot thorws exception."""
        fake_volume, fake_snapshot = _stub_volume(), _stub_snapshot()
        mock_mdp.side_effect = exception.UnableToExecuteHyperScaleCmd(
            cmd_out='mock error')
        self.assertRaises(exception.UnableToExecuteHyperScaleCmd,
                          self.driver.create_volume_from_snapshot, fake_volume,
                          fake_snapshot)

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale'
                '.HyperScaleDriver._select_rt')
    def test_create_volume_from_snapshot_with_no_replica(self, mock_srt,
                                                         mock_mdp):
        """Test case create volume from snapshot when there is no replica."""
        mock_obj = {'payload': {}}
        mock_mdp.return_value = (mock_obj, None)
        mock_srt.return_value = (None, None)
        fake_volume, fake_snapshot = _stub_volume(), _stub_snapshot()
        self.assertDictContainsSubset({
            'provider_location': 'hyperscale-sv:/hyperscale'
        }, self.driver.create_volume_from_snapshot(fake_volume, fake_snapshot))

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale'
                '.HyperScaleDriver._select_rt')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata_value')
    def test_create_volume_from_snapshot_with_replica(self, mock_gvmv,
                                                      mock_srt, mock_mdp):
        """Test case create volume from snapshot when there is replica."""
        mock_gvmv.side_effect = VRTSHyperScaleDriverTestCase.gvmv_side_effect
        mock_obj = {'payload': {}}
        mock_mdp.return_value = (mock_obj, None)
        mock_srt.return_value = ('{1234}', '192.0.2.2')
        fake_volume, fake_snapshot = _stub_volume(), _stub_snapshot()
        metadata = {
            'current_dn_owner': '{1234}',
            'Potential_secondary_key': '{1234}',
            'Primary_datanode_ip': '192.0.2.1',
            'Potential_secondary_ip': '192.0.2.2',
            'current_dn_ip': '192.0.2.1',
            'snapshot_id': fake_snapshot['id'],
            'parent_volume_guid': '{' + fake_snapshot['volume']['id'] + '}'
        }
        self.assertDictContainsSubset({
            'provider_location': 'hyperscale-sv:/hyperscale',
            'metadata': metadata
        }, self.driver.create_volume_from_snapshot(fake_volume, fake_snapshot))

    def test_initialize_connection(self):
        """Test case intialize_connection."""
        fake_volume = _stub_volume()
        expected_data = {
            'driver_volume_type': 'veritas_hyperscale',
            'data': {
                'export': fake_volume['provider_location'],
                'name': fake_volume['name']
            }
        }
        self.assertEqual(expected_data,
                         self.driver.initialize_connection(fake_volume, None))

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_compute_plane')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.episodic_snap')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata_value')
    def test_create_snapshot_with_exception(
            self, mock_gvmv, mock_es, mock_mcp):
        """Test case create snapshot throws exception."""
        mock_gvmv.side_effect = VRTSHyperScaleDriverTestCase.gvmv_side_effect
        mock_es_obj = {'payload': {'update': False}}
        mock_es.return_value = mock_es_obj
        mock_mcp.side_effect = exception.UnableToExecuteHyperScaleCmd(
            cmd_out='mock error')
        fake_snapshot = _stub_snapshot()
        self.assertRaises(exception.UnableToExecuteHyperScaleCmd,
                          self.driver.create_snapshot, fake_snapshot)

    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_controller')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_data_plane')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.episodic_snap')
    @mock.patch('cinder.volume.drivers.veritas.vrtshyperscale.HyperScaleDriver'
                '._get_volume_metadata_value')
    @mock.patch('cinder.volume.drivers.veritas.utils'
                '.message_compute_plane')
    def test_create_snapshot_user(
            self, mock_cdp, mock_gvmv, mock_es, mock_mdp, mock_mc):
        """Test case user snapshot."""
        mock_gvmv.side_effect = VRTSHyperScaleDriverTestCase.gvmv_side_effect
        mock_es_obj = {'payload': {'update': False}}
        mock_es.return_value = mock_es_obj
        mock_obj = {'payload': {}}
        mock_mdp.return_value = ("", None)
        mock_mc.return_value = ("", None)
        mock_cdp.return_value = (mock_obj, None)
        fake_snapshot = _stub_snapshot()
        expected = {
            'metadata': {
                'status': 'creating',
                'datanode_ip': '192.0.2.1',
                'TYPE': vrts.TYPE_USER_SNAP
            }
        }
        self.assertEqual(expected, self.driver.create_snapshot(fake_snapshot))



#   Copyright 2013, Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import mock
from oslo_serialization import jsonutils
from six.moves import http_client
import webob

from cinder import context
from cinder import db
from cinder.objects import fields
from cinder import test
from cinder.tests.unit.api import fakes
from cinder.tests.unit.api.v2 import fakes as v2_fakes
from cinder.tests.unit import fake_constants as fake


def fake_snapshot_get(context, snapshot_id):
    snapshot = v2_fakes.fake_snapshot(snapshot_id)

    if snapshot_id == fake.SNAPSHOT_ID:
        snapshot['status'] = fields.SnapshotStatus.CREATING
    else:
        snapshot['status'] = fields.SnapshotStatus.ERROR
    return snapshot


class SnapshotActionsTest(test.TestCase):

    def setUp(self):
        super(SnapshotActionsTest, self).setUp()
        self.user_ctxt = context.RequestContext(
            fake.USER_ID, fake.PROJECT_ID, auth_token=True)

    @mock.patch('cinder.db.snapshot_update', autospec=True)
    @mock.patch('cinder.db.sqlalchemy.api._snapshot_get',
                side_effect=fake_snapshot_get)
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    def test_update_snapshot_status(self, metadata_get, *args):

        body = {'os-update_snapshot_status':
                {'status': fields.SnapshotStatus.AVAILABLE}}
        req = webob.Request.blank('/v2/%s/snapshots/%s/action' % (
            fake.PROJECT_ID, fake.SNAPSHOT_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        self.assertEqual(http_client.ACCEPTED, res.status_int)

    @mock.patch('cinder.db.sqlalchemy.api._snapshot_get',
                side_effect=fake_snapshot_get)
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    def test_update_snapshot_status_invalid_status(self, metadata_get, *args):
        body = {'os-update_snapshot_status': {'status': 'in-use'}}
        req = webob.Request.blank('/v2/%s/snapshots/%s/action' % (
            fake.PROJECT_ID, fake.SNAPSHOT_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        self.assertEqual(http_client.BAD_REQUEST, res.status_int)

    def test_update_snapshot_status_without_status(self):
        self.mock_object(db, 'snapshot_get', fake_snapshot_get)
        body = {'os-update_snapshot_status': {}}
        req = webob.Request.blank('/v2/%s/snapshots/%s/action' % (
            fake.PROJECT_ID, fake.SNAPSHOT_ID))
        req.method = "POST"
        req.body = jsonutils.dump_as_bytes(body)
        req.headers["content-type"] = "application/json"

        res = req.get_response(fakes.wsgi_app(
            fake_auth_context=self.user_ctxt))
        self.assertEqual(http_client.BAD_REQUEST, res.status_int)



# Copyright (c) 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64
import binascii
import random
import string

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import excutils
from oslo_utils import units
import six

from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder.objects import fields
from cinder.volume import configuration
from cinder.volume.drivers.coprhd.helpers import (
    authentication as coprhd_auth)
from cinder.volume.drivers.coprhd.helpers import (
    commoncoprhdapi as coprhd_utils)
from cinder.volume.drivers.coprhd.helpers import (
    consistencygroup as coprhd_cg)
from cinder.volume.drivers.coprhd.helpers import exportgroup as coprhd_eg
from cinder.volume.drivers.coprhd.helpers import host as coprhd_host
from cinder.volume.drivers.coprhd.helpers import snapshot as coprhd_snap
from cinder.volume.drivers.coprhd.helpers import tag as coprhd_tag

from cinder.volume.drivers.coprhd.helpers import (
    virtualarray as coprhd_varray)
from cinder.volume.drivers.coprhd.helpers import volume as coprhd_vol
from cinder.volume import utils as volume_utils
from cinder.volume import volume_types

LOG = logging.getLogger(__name__)

MAX_RETRIES = 10
INTERVAL_10_SEC = 10

volume_opts = [
    cfg.StrOpt('coprhd_hostname',
               default=None,
               help='Hostname for the CoprHD Instance'),
    cfg.PortOpt('coprhd_port',
                default=4443,
                help='Port for the CoprHD Instance'),
    cfg.StrOpt('coprhd_username',
               default=None,
               help='Username for accessing the CoprHD Instance'),
    cfg.StrOpt('coprhd_password',
               default=None,
               help='Password for accessing the CoprHD Instance',
               secret=True),
    cfg.StrOpt('coprhd_tenant',
               default=None,
               help='Tenant to utilize within the CoprHD Instance'),
    cfg.StrOpt('coprhd_project',
               default=None,
               help='Project to utilize within the CoprHD Instance'),
    cfg.StrOpt('coprhd_varray',
               default=None,
               help='Virtual Array to utilize within the CoprHD Instance'),
    cfg.BoolOpt('coprhd_emulate_snapshot',
                default=False,
                help='True | False to indicate if the storage array '
                'in CoprHD is VMAX or VPLEX')
]

CONF = cfg.CONF
CONF.register_opts(volume_opts, group=configuration.SHARED_CONF_GROUP)

URI_VPOOL_VARRAY_CAPACITY = '/block/vpools/{0}/varrays/{1}/capacity'
URI_BLOCK_EXPORTS_FOR_INITIATORS = '/block/exports?initiators={0}'
EXPORT_RETRY_COUNT = 5
MAX_DEFAULT_NAME_LENGTH = 128
MAX_SNAPSHOT_NAME_LENGTH = 63
MAX_CONSISTENCY_GROUP_NAME_LENGTH = 64
MAX_SIO_LEN = 31


def retry_wrapper(func):
    def try_and_retry(*args, **kwargs):
        retry = False
        try:
            return func(*args, **kwargs)
        except coprhd_utils.CoprHdError as e:
            # if we got an http error and
            # the string contains 401 or if the string contains the word cookie
            if (e.err_code == coprhd_utils.CoprHdError.HTTP_ERR and
                (e.msg.find('401') != -1 or
                 e.msg.lower().find('cookie') != -1)):
                retry = True
                args[0].AUTHENTICATED = False
            else:
                exception_message = (_("\nCoprHD Exception: %(msg)s\n") %
                                     {'msg': e.msg})
                LOG.exception(exception_message)
                raise exception.VolumeBackendAPIException(
                    data=exception_message)
        except Exception as exc:
            exception_message = (_("\nGeneral Exception: %(exec_info)s\n") %
                                 {'exec_info':
                                  encodeutils.exception_to_unicode(exc)})
            LOG.exception(exception_message)
            raise exception.VolumeBackendAPIException(
                data=exception_message)

        if retry:
            return func(*args, **kwargs)

    return try_and_retry


class EMCCoprHDDriverCommon(object):

    OPENSTACK_TAG = 'OpenStack'

    def __init__(self, protocol, default_backend_name, configuration=None):
        self.AUTHENTICATED = False
        self.protocol = protocol
        self.configuration = configuration
        self.configuration.append_config_values(volume_opts)

        self.init_coprhd_api_components()

        self.stats = {'driver_version': '3.0.0.0',
                      'free_capacity_gb': 'unknown',
                      'reserved_percentage': '0',
                      'storage_protocol': protocol,
                      'total_capacity_gb': 'unknown',
                      'vendor_name': 'CoprHD',
                      'volume_backend_name':
                      self.configuration.volume_backend_name or
                      default_backend_name}

    def init_coprhd_api_components(self):

        coprhd_utils.AUTH_TOKEN = None

        # instantiate coprhd api objects for later use
        self.volume_obj = coprhd_vol.Volume(
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port)

        self.exportgroup_obj = coprhd_eg.ExportGroup(
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port)

        self.host_obj = coprhd_host.Host(
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port)

        self.varray_obj = coprhd_varray.VirtualArray(
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port)

        self.snapshot_obj = coprhd_snap.Snapshot(
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port)

        self.consistencygroup_obj = coprhd_cg.ConsistencyGroup(
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port)

        self.tag_obj = coprhd_tag.Tag(
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port)

    def check_for_setup_error(self):
        # validate all of the coprhd_* configuration values
        if self.configuration.coprhd_hostname is None:
            message = _("coprhd_hostname is not set in cinder configuration")
            raise exception.VolumeBackendAPIException(data=message)

        if self.configuration.coprhd_port is None:
            message = _("coprhd_port is not set in cinder configuration")
            raise exception.VolumeBackendAPIException(data=message)

        if self.configuration.coprhd_username is None:
            message = _("coprhd_username is not set in cinder configuration")
            raise exception.VolumeBackendAPIException(data=message)

        if self.configuration.coprhd_password is None:
            message = _("coprhd_password is not set in cinder configuration")
            raise exception.VolumeBackendAPIException(data=message)

        if self.configuration.coprhd_tenant is None:
            message = _("coprhd_tenant is not set in cinder configuration")
            raise exception.VolumeBackendAPIException(data=message)

        if self.configuration.coprhd_project is None:
            message = _("coprhd_project is not set in cinder configuration")
            raise exception.VolumeBackendAPIException(data=message)

        if self.configuration.coprhd_varray is None:
            message = _("coprhd_varray is not set in cinder configuration")
            raise exception.VolumeBackendAPIException(data=message)

    def authenticate_user(self):
        # we should check to see if we are already authenticated before blindly
        # doing it again
        if self.AUTHENTICATED is False:
            obj = coprhd_auth.Authentication(
                self.configuration.coprhd_hostname,
                self.configuration.coprhd_port)

            username = self.configuration.coprhd_username
            password = self.configuration.coprhd_password

            coprhd_utils.AUTH_TOKEN = obj.authenticate_user(username,
                                                            password)
            self.AUTHENTICATED = True

    def create_volume(self, vol, driver, truncate_name=False):
        self.authenticate_user()
        name = self._get_resource_name(vol, MAX_DEFAULT_NAME_LENGTH,
                                       truncate_name)
        size = int(vol.size) * units.Gi

        vpool = self._get_vpool(vol)
        self.vpool = vpool['CoprHD:VPOOL']

        try:
            coprhd_cgid = None
            try:
                if vol.group_id:
                    if volume_utils.is_group_a_cg_snapshot_type(vol.group):
                        coprhd_cgid = self._get_coprhd_cgid(vol.group_id)
            except KeyError:
                coprhd_cgid = None
            except AttributeError:
                coprhd_cgid = None

            full_project_name = ("%s/%s" % (self.configuration.coprhd_tenant,
                                            self.configuration.coprhd_project)
                                 )
            self.volume_obj.create(full_project_name, name, size,
                                   self.configuration.coprhd_varray,
                                   self.vpool,
                                   # no longer specified in volume creation
                                   sync=True,
                                   # no longer specified in volume creation
                                   consistencygroup=coprhd_cgid)

        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Volume %(name)s: create failed\n%(err)s") %
                              {'name': name, 'err': six.text_type(e.msg)})

            log_err_msg = ("Volume : %s creation failed" % name)
            self._raise_or_log_exception(
                e.err_code, coprhd_err_msg, log_err_msg)

    @retry_wrapper
    def create_consistencygroup(self, context, group, truncate_name=False):
        self.authenticate_user()
        name = self._get_resource_name(group,
                                       MAX_CONSISTENCY_GROUP_NAME_LENGTH,
                                       truncate_name)

        try:
            self.consistencygroup_obj.create(
                name,
                self.configuration.coprhd_project,
                self.configuration.coprhd_tenant)

            cg_uri = self.consistencygroup_obj.consistencygroup_query(
                name,
                self.configuration.coprhd_project,
                self.configuration.coprhd_tenant)

            self.set_tags_for_resource(
                coprhd_cg.ConsistencyGroup.URI_CONSISTENCY_GROUP_TAGS,
                cg_uri, group)

        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Consistency Group %(name)s:"
                                " create failed\n%(err)s") %
                              {'name': name, 'err': six.text_type(e.msg)})

            log_err_msg = ("Consistency Group : %s creation failed" %
                           name)
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)

    @retry_wrapper
    def update_consistencygroup(self, group, add_volumes,
                                remove_volumes):
        self.authenticate_user()
        model_update = {'status': fields.GroupStatus.AVAILABLE}
        cg_uri = self._get_coprhd_cgid(group.id)
        add_volnames = []
        remove_volnames = []

        try:
            if add_volumes:
                for vol in add_volumes:
                    vol_name = self._get_coprhd_volume_name(vol)
                    add_volnames.append(vol_name)

            if remove_volumes:
                for vol in remove_volumes:
                    vol_name = self._get_coprhd_volume_name(vol)
                    remove_volnames.append(vol_name)

            self.consistencygroup_obj.update(
                cg_uri,
                self.configuration.coprhd_project,
                self.configuration.coprhd_tenant,
                add_volnames, remove_volnames, True)

            return model_update, None, None

        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Consistency Group %(cg_uri)s:"
                                " update failed\n%(err)s") %
                              {'cg_uri': cg_uri, 'err': six.text_type(e.msg)})

            log_err_msg = ("Consistency Group : %s update failed" %
                           cg_uri)
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)

    @retry_wrapper
    def delete_consistencygroup(self, context, group, volumes,
                                truncate_name=False):
        self.authenticate_user()
        name = self._get_resource_name(group,
                                       MAX_CONSISTENCY_GROUP_NAME_LENGTH,
                                       truncate_name)
        volumes_model_update = []

        try:
            for vol in volumes:
                try:
                    vol_name = self._get_coprhd_volume_name(vol)
                    full_project_name = "%s/%s" % (
                        self.configuration.coprhd_tenant,
                        self.configuration.coprhd_project)

                    self.volume_obj.delete(full_project_name, vol_name,
                                           sync=True,
                                           force_delete=True)

                    update_item = {'id': vol.id,
                                   'status':
                                   fields.GroupStatus.DELETED}
                    volumes_model_update.append(update_item)

                except exception.VolumeBackendAPIException:
                    update_item = {'id': vol.id,
                                   'status': fields.ConsistencyGroupStatus.
                                   ERROR_DELETING}

                    volumes_model_update.append(update_item)

                    LOG.exception("Failed to delete the volume %s of CG.",
                                  vol.name)

            self.consistencygroup_obj.delete(
                name,
                self.configuration.coprhd_project,
                self.configuration.coprhd_tenant)

            model_update = {}
            model_update['status'] = group.status

            return model_update, volumes_model_update

        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Consistency Group %(name)s:"
                                " delete failed\n%(err)s") %
                              {'name': name, 'err': six.text_type(e.msg)})

            log_err_msg = ("Consistency Group : %s deletion failed" %
                           name)
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)

    @retry_wrapper
    def create_cgsnapshot(self, cgsnapshot, snapshots, truncate_name=False):
        self.authenticate_user()

        snapshots_model_update = []
        cgsnapshot_name = self._get_resource_name(cgsnapshot,
                                                  MAX_SNAPSHOT_NAME_LENGTH,
                                                  truncate_name)

        cg_id = None
        cg_group = None

        try:
            cg_id = cgsnapshot.group_id
            cg_group = cgsnapshot.group
        except AttributeError:
            pass

        cg_name = None
        coprhd_cgid = None

        if cg_id:
            coprhd_cgid = self._get_coprhd_cgid(cg_id)
            cg_name = self._get_consistencygroup_name(cg_group)

        LOG.info('Start to create cgsnapshot for consistency group'
                 ': %(group_name)s',
                 {'group_name': cg_name})

        try:
            self.snapshot_obj.snapshot_create(
                'block',
                'consistency-groups',
                coprhd_cgid,
                cgsnapshot_name,
                False,
                True)

            for snapshot in snapshots:
                vol_id_of_snap = snapshot.volume_id

                # Finding the volume in CoprHD for this volume id
                tagname = "OpenStack:id:" + vol_id_of_snap
                rslt = coprhd_utils.search_by_tag(
                    coprhd_vol.Volume.URI_SEARCH_VOLUMES_BY_TAG.format(
                        tagname),
                    self.configuration.coprhd_hostname,
                    self.configuration.coprhd_port)

                if not rslt:
                    continue

                vol_uri = rslt[0]

                snapshots_of_volume = self.snapshot_obj.snapshot_list_uri(
                    'block',
                    'volumes',
                    vol_uri)

                for snapUri in snapshots_of_volume:
                    snapshot_obj = self.snapshot_obj.snapshot_show_uri(
                        'block',
                        vol_uri,
                        snapUri['id'])

                    if not coprhd_utils.get_node_value(snapshot_obj,
                                                       'inactive'):

                        # Creating snapshot for a consistency group.
                        # When we create a consistency group snapshot on
                        # coprhd then each snapshot of volume in the
                        # consistencygroup will be given a subscript. Ex if
                        # the snapshot name is cgsnap1 and lets say there are
                        # three vols(a,b,c) in CG. Then the names of snapshots
                        # of the volumes in cg on coprhd end will be like
                        # cgsnap1-1 cgsnap1-2 cgsnap1-3. So, we list the
                        # snapshots of the volume under consideration and then
                        # split the name  using - from the ending as prefix
                        # and postfix. We compare the prefix to the cgsnapshot
                        # name and filter our the snapshots that correspond to
                        # the cgsnapshot

                        if '-' in snapshot_obj['name']:
                            (prefix, postfix) = snapshot_obj[
                                'name'].rsplit('-', 1)

                            if cgsnapshot_name == prefix:
                                self.set_tags_for_resource(
                                    coprhd_snap.Snapshot.
                                    URI_BLOCK_SNAPSHOTS_TAG,
                                    snapUri['id'],
                                    snapshot)

                        elif cgsnapshot_name == snapshot_obj['name']:
                            self.set_tags_for_resource(
                                coprhd_snap.Snapshot.URI_BLOCK_SNAPSHOTS_TAG,
                                snapUri['id'],
                                snapshot)

                snapshot['status'] = fields.SnapshotStatus.AVAILABLE
                snapshots_model_update.append(
                    {'id': snapshot.id, 'status':
                     fields.SnapshotStatus.AVAILABLE})

            model_update = {'status': fields.ConsistencyGroupStatus.AVAILABLE}

            return model_update, snapshots_model_update

        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Snapshot for Consistency Group %(cg_name)s:"
                                " create failed\n%(err)s") %
                              {'cg_name': cg_name,
                               'err': six.text_type(e.msg)})

            log_err_msg = ("Snapshot %(name)s for Consistency"
                           " Group: %(cg_name)s creation failed" %
                           {'cg_name': cg_name,
                            'name': cgsnapshot_name})
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)

    @retry_wrapper
    def delete_cgsnapshot(self, cgsnapshot, snapshots, truncate_name=False):
        self.authenticate_user()
        cgsnapshot_id = cgsnapshot.id
        cgsnapshot_name = self._get_resource_name(cgsnapshot,
                                                  MAX_SNAPSHOT_NAME_LENGTH,
                                                  truncate_name)

        snapshots_model_update = []

        cg_id = None
        cg_group = None

        try:
            cg_id = cgsnapshot.group_id
            cg_group = cgsnapshot.group
        except AttributeError:
            pass

        coprhd_cgid = self._get_coprhd_cgid(cg_id)
        cg_name = self._get_consistencygroup_name(cg_group)

        model_update = {}
        LOG.info('Delete cgsnapshot %(snap_name)s for consistency group: '
                 '%(group_name)s', {'snap_name': cgsnapshot.name,
                                    'group_name': cg_name})

        try:
            uri = None
            try:
                uri = self.snapshot_obj.snapshot_query('block',
                                                       'consistency-groups',
                                                       coprhd_cgid,
                                                       cgsnapshot_name + '-1')
            except coprhd_utils.CoprHdError as e:
                if e.err_code == coprhd_utils.CoprHdError.NOT_FOUND_ERR:
                    uri = self.snapshot_obj.snapshot_query(
                        'block',
                        'consistency-groups',
                        coprhd_cgid,
                        cgsnapshot_name)
            self.snapshot_obj.snapshot_delete_uri(
                'block',
                coprhd_cgid,
                uri,
                True,
                0)

            for snapshot in snapshots:
                snapshots_model_update.append(
                    {'id': snapshot.id,
                     'status': fields.SnapshotStatus.DELETED})

            return model_update, snapshots_model_update

        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Snapshot %(cgsnapshot_id)s: for"
                                " Consistency Group %(cg_name)s: delete"
                                " failed\n%(err)s") %
                              {'cgsnapshot_id': cgsnapshot_id,
                               'cg_name': cg_name,
                               'err': six.text_type(e.msg)})

            log_err_msg = ("Snapshot %(name)s for Consistency"
                           " Group: %(cg_name)s deletion failed" %
                           {'cg_name': cg_name,
                            'name': cgsnapshot_name})
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)

    @retry_wrapper
    def set_volume_tags(self, vol, exempt_tags=None, truncate_name=False):
        if exempt_tags is None:
            exempt_tags = []

        self.authenticate_user()
        name = self._get_resource_name(vol,
                                       MAX_DEFAULT_NAME_LENGTH,
                                       truncate_name)
        full_project_name = ("%s/%s" % (
            self.configuration.coprhd_tenant,
            self.configuration.coprhd_project))

        vol_uri = self.volume_obj.volume_query(full_project_name,
                                               name)

        self.set_tags_for_resource(
            coprhd_vol.Volume.URI_TAG_VOLUME, vol_uri, vol, exempt_tags)

    @retry_wrapper
    def set_tags_for_resource(self, uri, resource_id, resource,
                              exempt_tags=None):
        if exempt_tags is None:
            exempt_tags = []

        self.authenticate_user()

        # first, get the current tags that start with the OPENSTACK_TAG
        # eyecatcher
        formattedUri = uri.format(resource_id)
        remove_tags = []
        currentTags = self.tag_obj.list_tags(formattedUri)
        for cTag in currentTags:
            if cTag.startswith(self.OPENSTACK_TAG):
                remove_tags.append(cTag)

        try:
            if remove_tags:
                self.tag_obj.tag_resource(uri,
                                          resource_id,
                                          None,
                                          remove_tags)
        except coprhd_utils.CoprHdError as e:
            if e.err_code == coprhd_utils.CoprHdError.SOS_FAILURE_ERR:
                LOG.debug("CoprHdError adding the tag:\n %s", e.msg)

        # now add the tags for the resource
        add_tags = []
        # put all the openstack resource properties into the CoprHD resource

        try:
            for prop, value in vars(resource).items():
                try:
                    if prop in exempt_tags:
                        continue

                    if prop.startswith("_"):
                        prop = prop.replace("_", '', 1)

                    # don't put the status in, it's always the status before
                    # the current transaction
                    if ((not prop.startswith("status") and not
                         prop.startswith("obj_status") and
                         prop != "obj_volume") and value):
                        tag = ("%s:%s:%s" %
                               (self.OPENSTACK_TAG, prop,
                                six.text_type(value)))

                        if len(tag) > 128:
                            tag = tag[0:128]
                        add_tags.append(tag)
                except TypeError:
                    LOG.error(
                        "Error tagging the resource property %s", prop)
        except TypeError:
            LOG.error("Error tagging the resource properties")

        try:
            self.tag_obj.tag_resource(
                uri,
                resource_id,
                add_tags,
                None)
        except coprhd_utils.CoprHdError as e:
            if e.err_code == coprhd_utils.CoprHdError.SOS_FAILURE_ERR:
                LOG.debug(
                    "Adding the tag failed. CoprHdError: %s", e.msg)

        return self.tag_obj.list_tags(formattedUri)

    @retry_wrapper
    def create_cloned_volume(self, vol, src_vref, truncate_name=False):
        """Creates a clone of the specified volume."""
        self.authenticate_user()
        name = self._get_resource_name(vol,
                                       MAX_DEFAULT_NAME_LENGTH,
                                       truncate_name)
        srcname = self._get_coprhd_volume_name(src_vref)

        try:
            if src_vref.group_id:
                raise coprhd_utils.CoprHdError(
                    coprhd_utils.CoprHdError.SOS_FAILURE_ERR,
                    _("Clone can't be taken individually on a volume"
                      " that is part of a Consistency Group"))
        except KeyError as e:
            pass
        except AttributeError:
            pass
        try:
            (storageres_type,
             storageres_typename) = self.volume_obj.get_storageAttributes(
                srcname, None, None)

            resource_id = self.volume_obj.storage_resource_query(
                storageres_type,
                srcname,
                None,
                None,
                self.configuration.coprhd_project,
                self.configuration.coprhd_tenant)

            self.volume_obj.clone(
                name,
                resource_id,
                sync=True)

            full_project_name = "%s/%s" % (
                self.configuration.coprhd_tenant,
                self.configuration.coprhd_project)

            detachable = self.volume_obj.is_volume_detachable(
                full_project_name, name)
            LOG.debug("Is volume detachable : %s", detachable)

            # detach it from the source volume immediately after creation
            if detachable:
                self.volume_obj.volume_clone_detach(
                    "", full_project_name, name, True)

        except IndexError:
            LOG.exception("Volume clone detach returned empty task list")

        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Volume %(name)s: clone failed\n%(err)s") %
                              {'name': name, 'err': six.text_type(e.msg)})

            log_err_msg = ("Volume : {%s} clone failed" % name)
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)

        src_vol_size = 0
        dest_vol_size = 0

        try:
            src_vol_size = src_vref.size
        except AttributeError:
            src_vol_size = src_vref.volume_size

        try:
            dest_vol_size = vol.size
        except AttributeError:
            dest_vol_size = vol.volume_size

        if dest_vol_size > src_vol_size:
            size_in_bytes = coprhd_utils.to_bytes("%sG" % dest_vol_size)
            try:
                self.volume_obj.expand(
                    ("%s/%s" % (self.configuration.coprhd_tenant,
                                self.configuration.coprhd_project)), name,
                    size_in_bytes,
                    True)
            except coprhd_utils.CoprHdError as e:
                coprhd_err_msg = (_("Volume %(volume_name)s: expand failed"
                                    "\n%(err)s") %
                                  {'volume_name': name,
                                   'err': six.text_type(e.msg)})

                log_err_msg = ("Volume : %s expand failed" % name)
                self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                             log_err_msg)

    @retry_wrapper
    def expand_volume(self, vol, new_size):
        """expands the volume to new_size specified."""
        self.authenticate_user()
        volume_name = self._get_coprhd_volume_name(vol)
        size_in_bytes = coprhd_utils.to_bytes("%sG" % new_size)

        try:
            self.volume_obj.expand(
                ("%s/%s" % (self.configuration.coprhd_tenant,
                            self.configuration.coprhd_project)), volume_name,
                size_in_bytes,
                True)
        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Volume %(volume_name)s:"
                                " expand failed\n%(err)s") %
                              {'volume_name': volume_name,
                               'err': six.text_type(e.msg)})

            log_err_msg = ("Volume : %s expand failed" %
                           volume_name)
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)

    @retry_wrapper
    def create_volume_from_snapshot(self, snapshot, volume,
                                    truncate_name=False):
        """Creates volume from given snapshot ( snapshot clone to volume )."""
        self.authenticate_user()

        if self.configuration.coprhd_emulate_snapshot:
            self.create_cloned_volume(volume, snapshot, truncate_name)
            return

        try:
            if snapshot.group_snapshot_id:
                raise coprhd_utils.CoprHdError(
                    coprhd_utils.CoprHdError.SOS_FAILURE_ERR,
                    _("Volume cannot be created individually from a snapshot "
                      "that is part of a Consistency Group"))
        except AttributeError:
            pass

        src_snapshot_name = None
        src_vol_ref = snapshot.volume
        new_volume_name = self._get_resource_name(volume,
                                                  MAX_DEFAULT_NAME_LENGTH,
                                                  truncate_name)

        try:
            coprhd_vol_info = self._get_coprhd_volume_name(
                src_vol_ref, True)
            src_snapshot_name = self._get_coprhd_snapshot_name(
                snapshot, coprhd_vol_info['volume_uri'])

            (storageres_type,
             storageres_typename) = self.volume_obj.get_storageAttributes(
                coprhd_vol_info['volume_name'], None, src_snapshot_name)

            resource_id = self.volume_obj.storage_resource_query(
                storageres_type,
                coprhd_vol_info['volume_name'],
                None,
                src_snapshot_name,
                self.configuration.coprhd_project,
                self.configuration.coprhd_tenant)

            self.volume_obj.clone(
                new_volume_name,
                resource_id,
                sync=True)

        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Snapshot %(src_snapshot_name)s:"
                                " clone failed\n%(err)s") %
                              {'src_snapshot_name': src_snapshot_name,
                               'err': six.text_type(e.msg)})

            log_err_msg = ("Snapshot : %s clone failed" %
                           src_snapshot_name)
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)

        if volume.size > snapshot.volume_size:
            size_in_bytes = coprhd_utils.to_bytes("%sG" % volume.size)

            try:
                self.volume_obj.expand(
                    ("%s/%s" % (self.configuration.coprhd_tenant,
                                self.configuration.coprhd_project)),
                    new_volume_name, size_in_bytes, True)

            except coprhd_utils.CoprHdError as e:
                coprhd_err_msg = (_("Volume %(volume_name)s: expand failed"
                                    "\n%(err)s") %
                                  {'volume_name': new_volume_name,
                                   'err': six.text_type(e.msg)})

                log_err_msg = ("Volume : %s expand failed" %
                               new_volume_name)
                self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                             log_err_msg)

    @retry_wrapper
    def delete_volume(self, vol):
        self.authenticate_user()
        name = self._get_coprhd_volume_name(vol)
        try:
            full_project_name = ("%s/%s" % (
                self.configuration.coprhd_tenant,
                self.configuration.coprhd_project))
            self.volume_obj.delete(full_project_name, name, sync=True)
        except coprhd_utils.CoprHdError as e:
            if e.err_code == coprhd_utils.CoprHdError.NOT_FOUND_ERR:
                LOG.info(
                    "Volume %s"
                    " no longer exists; volume deletion is"
                    " considered successful.", name)
            else:
                coprhd_err_msg = (_("Volume %(name)s: delete failed"
                                    "\n%(err)s") %
                                  {'name': name, 'err': six.text_type(e.msg)})

                log_err_msg = ("Volume : %s delete failed" % name)
                self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                             log_err_msg)

    @retry_wrapper
    def create_snapshot(self, snapshot, truncate_name=False):
        self.authenticate_user()

        volume = snapshot.volume

        try:
            if volume.group_id:
                raise coprhd_utils.CoprHdError(
                    coprhd_utils.CoprHdError.SOS_FAILURE_ERR,
                    _("Snapshot can't be taken individually on a volume"
                      " that is part of a Consistency Group"))
        except KeyError:
            LOG.info("No Consistency Group associated with the volume")

        if self.configuration.coprhd_emulate_snapshot:
            self.create_cloned_volume(snapshot, volume, truncate_name)
            self.set_volume_tags(
                snapshot, ['_volume', '_obj_volume_type'], truncate_name)
            return

        try:
            snapshotname = self._get_resource_name(snapshot,
                                                   MAX_SNAPSHOT_NAME_LENGTH,
                                                   truncate_name)
            vol = snapshot.volume

            volumename = self._get_coprhd_volume_name(vol)
            projectname = self.configuration.coprhd_project
            tenantname = self.configuration.coprhd_tenant
            storageres_type = 'block'
            storageres_typename = 'volumes'
            resource_uri = self.snapshot_obj.storage_resource_query(
                storageres_type,
                volume_name=volumename,
                cg_name=None,
                project=projectname,
                tenant=tenantname)
            inactive = False
            sync = True
            self.snapshot_obj.snapshot_create(
                storageres_type,
                storageres_typename,
                resource_uri,
                snapshotname,
                inactive,
                sync)

            snapshot_uri = self.snapshot_obj.snapshot_query(
                storageres_type,
                storageres_typename,
                resource_uri,
                snapshotname)

            self.set_tags_for_resource(
                coprhd_snap.Snapshot.URI_BLOCK_SNAPSHOTS_TAG,
                snapshot_uri, snapshot, ['_volume'])

        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Snapshot: %(snapshotname)s, create failed"
                                "\n%(err)s") % {'snapshotname': snapshotname,
                                                'err': six.text_type(e.msg)})

            log_err_msg = ("Snapshot : %s create failed" % snapshotname)
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)

    @retry_wrapper
    def delete_snapshot(self, snapshot):
        self.authenticate_user()

        vol = snapshot.volume

        try:
            if vol.group_id:
                raise coprhd_utils.CoprHdError(
                    coprhd_utils.CoprHdError.SOS_FAILURE_ERR,
                    _("Snapshot delete can't be done individually on a volume"
                      " that is part of a Consistency Group"))
        except KeyError:
            LOG.info("No Consistency Group associated with the volume")

        if self.configuration.coprhd_emulate_snapshot:
            self.delete_volume(snapshot)
            return

        snapshotname = None
        try:
            volumename = self._get_coprhd_volume_name(vol)
            projectname = self.configuration.coprhd_project
            tenantname = self.configuration.coprhd_tenant
            storageres_type = 'block'
            storageres_typename = 'volumes'
            resource_uri = self.snapshot_obj.storage_resource_query(
                storageres_type,
                volume_name=volumename,
                cg_name=None,
                project=projectname,
                tenant=tenantname)
            if resource_uri is None:
                LOG.info(
                    "Snapshot %s"
                    " is not found; snapshot deletion"
                    " is considered successful.", snapshotname)
            else:
                snapshotname = self._get_coprhd_snapshot_name(
                    snapshot, resource_uri)

                self.snapshot_obj.snapshot_delete(
                    storageres_type,
                    storageres_typename,
                    resource_uri,
                    snapshotname,
                    sync=True)
        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Snapshot %s : Delete Failed\n") %
                              snapshotname)

            log_err_msg = ("Snapshot : %s delete failed" % snapshotname)
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)

    @retry_wrapper
    def initialize_connection(self, volume, protocol, initiator_ports,
                              hostname):

        try:
            self.authenticate_user()
            volumename = self._get_coprhd_volume_name(volume)
            foundgroupname = self._find_exportgroup(initiator_ports)
            foundhostname = None
            if foundgroupname is None:
                for i in range(len(initiator_ports)):
                    # check if this initiator is contained in any CoprHD Host
                    # object
                    LOG.debug(
                        "checking for initiator port: %s", initiator_ports[i])
                    foundhostname = self._find_host(initiator_ports[i])

                    if foundhostname:
                        LOG.info("Found host %s", foundhostname)
                        break

                if not foundhostname:
                    LOG.error("Auto host creation not supported")
                # create an export group for this host
                foundgroupname = foundhostname + 'SG'
                # create a unique name
                foundgroupname = foundgroupname + '-' + ''.join(
                    random.choice(string.ascii_uppercase +
                                  string.digits)
                    for x in range(6))
                self.exportgroup_obj.exportgroup_create(
                    foundgroupname,
                    self.configuration.coprhd_project,
                    self.configuration.coprhd_tenant,
                    self.configuration.coprhd_varray,
                    'Host',
                    foundhostname)

            LOG.debug(
                "adding the volume to the exportgroup : %s", volumename)

            self.exportgroup_obj.exportgroup_add_volumes(
                True,
                foundgroupname,
                self.configuration.coprhd_tenant,
                None,
                None,
                None,
                self.configuration.coprhd_project,
                [volumename],
                None,
                None)

            return self._find_device_info(volume, initiator_ports)

        except coprhd_utils.CoprHdError as e:
            raise coprhd_utils.CoprHdError(
                coprhd_utils.CoprHdError.SOS_FAILURE_ERR,
                (_("Attach volume (%(name)s) to host"
                   " (%(hostname)s) initiator (%(initiatorport)s)"
                   " failed:\n%(err)s") %
                 {'name': self._get_coprhd_volume_name(
                     volume),
                  'hostname': hostname,
                  'initiatorport': initiator_ports[0],
                  'err': six.text_type(e.msg)})
            )

    @retry_wrapper
    def terminate_connection(self, volume, protocol, initiator_ports,
                             hostname):
        try:
            self.authenticate_user()
            volumename = self._get_coprhd_volume_name(volume)
            full_project_name = ("%s/%s" % (self.configuration.coprhd_tenant,
                                            self.configuration.coprhd_project))
            voldetails = self.volume_obj.show(full_project_name, volumename)
            volid = voldetails['id']

            # find the exportgroups
            exports = self.volume_obj.get_exports_by_uri(volid)
            exportgroups = set()
            itls = exports['itl']
            for itl in itls:
                itl_port = itl['initiator']['port']
                if itl_port in initiator_ports:
                    exportgroups.add(itl['export']['id'])

            for exportgroup in exportgroups:
                self.exportgroup_obj.exportgroup_remove_volumes_by_uri(
                    exportgroup,
                    volid,
                    True,
                    None,
                    None,
                    None,
                    None)
            else:
                LOG.info(
                    "No export group found for the host: %s"
                    "; this is considered already detached.", hostname)

            return itls

        except coprhd_utils.CoprHdError as e:
            raise coprhd_utils.CoprHdError(
                coprhd_utils.CoprHdError.SOS_FAILURE_ERR,
                (_("Detaching volume %(volumename)s from host"
                   " %(hostname)s failed: %(err)s") %
                 {'volumename': volumename,
                  'hostname': hostname,
                  'err': six.text_type(e.msg)})
            )

    @retry_wrapper
    def _find_device_info(self, volume, initiator_ports):
        """Returns device_info in list of itls having the matched initiator.

        (there could be multiple targets, hence a list):
                [
                 {
                  "hlu":9,
                  "initiator":{...,"port":"20:00:00:25:B5:49:00:22"},
                  "export":{...},
                  "device":{...,"wwn":"600601602B802D00B62236585D0BE311"},
                  "target":{...,"port":"50:06:01:6A:46:E0:72:EF"},
                  "san_zone_name":"..."
                 },
                 {
                  "hlu":9,
                  "initiator":{...,"port":"20:00:00:25:B5:49:00:22"},
                  "export":{...},
                  "device":{...,"wwn":"600601602B802D00B62236585D0BE311"},
                  "target":{...,"port":"50:06:01:62:46:E0:72:EF"},
                  "san_zone_name":"..."
                 }
                ]
        """
        volumename = self._get_coprhd_volume_name(volume)
        full_project_name = ("%s/%s" % (self.configuration.coprhd_tenant,
                                        self.configuration.coprhd_project))
        vol_uri = self.volume_obj.volume_query(full_project_name, volumename)

        # The itl info shall be available at the first try since now export is
        # a synchronous call.  We are trying a few more times to accommodate
        # any delay on filling in the itl info after the export task is
        # completed.

        itls = []
        for x in range(MAX_RETRIES):
            exports = self.volume_obj.get_exports_by_uri(vol_uri)
            LOG.debug("Volume exports: ")
            LOG.info(vol_uri)
            LOG.debug(exports)
            for itl in exports['itl']:
                itl_port = itl['initiator']['port']
                if itl_port in initiator_ports:
                    found_device_number = itl['hlu']
                    if (found_device_number is not None and
                            found_device_number != '-1'):
                        # 0 is a valid number for found_device_number.
                        # Only loop if it is None or -1
                        LOG.debug("Found Device Number: %s",
                                  found_device_number)
                        itls.append(itl)

            if itls:
                break
            else:
                LOG.debug("Device Number not found yet."
                          " Retrying after 10 seconds...")
                eventlet.sleep(INTERVAL_10_SEC)

        if itls is None:
            # No device number found after 10 tries; return an empty itl
            LOG.info(
                "No device number has been found after 10 tries; "
                "this likely indicates an unsuccessful attach of "
                "volume volumename=%(volumename)s to"
                " initiator  initiator_ports=%(initiator_ports)s",
                {'volumename': volumename,
                    'initiator_ports': initiator_ports})

        return itls

    def _get_coprhd_cgid(self, cgid):
        tagname = self.OPENSTACK_TAG + ":id:" + cgid
        rslt = coprhd_utils.search_by_tag(
            coprhd_cg.ConsistencyGroup.URI_SEARCH_CONSISTENCY_GROUPS_BY_TAG.
            format(tagname),
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port)

        # if the result is empty, then search with the tagname as
        # "OpenStack:obj_id" the openstack attribute for id can be obj_id
        # instead of id. this depends on the version
        if rslt is None or len(rslt) == 0:
            tagname = self.OPENSTACK_TAG + ":obj_id:" + cgid
            rslt = coprhd_utils.search_by_tag(
                coprhd_cg.ConsistencyGroup
                .URI_SEARCH_CONSISTENCY_GROUPS_BY_TAG.
                format(tagname),
                self.configuration.coprhd_hostname,
                self.configuration.coprhd_port)

        if len(rslt) > 0:
            rslt_cg = self.consistencygroup_obj.show(
                rslt[0],
                self.configuration.coprhd_project,
                self.configuration.coprhd_tenant)
            return rslt_cg['id']
        else:
            raise coprhd_utils.CoprHdError(
                coprhd_utils.CoprHdError.NOT_FOUND_ERR,
                (_("Consistency Group %s not found") % cgid))

    def _get_consistencygroup_name(self, consisgrp):
        return consisgrp.name

    def _get_coprhd_snapshot_name(self, snapshot, resUri):
        tagname = self.OPENSTACK_TAG + ":id:" + snapshot['id']
        rslt = coprhd_utils.search_by_tag(
            coprhd_snap.Snapshot.URI_SEARCH_SNAPSHOT_BY_TAG.format(tagname),
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port)

        # if the result is empty, then search with the tagname
        # as "OpenStack:obj_id"
        # as snapshots will be having the obj_id instead of just id.
        if not rslt:
            tagname = self.OPENSTACK_TAG + ":obj_id:" + snapshot['id']
            rslt = coprhd_utils.search_by_tag(
                coprhd_snap.Snapshot.URI_SEARCH_SNAPSHOT_BY_TAG.format(
                    tagname),
                self.configuration.coprhd_hostname,
                self.configuration.coprhd_port)

        if rslt is None or len(rslt) == 0:
            return snapshot['name']
        else:
            rslt_snap = self.snapshot_obj.snapshot_show_uri(
                'block',
                resUri,
                rslt[0])
            return rslt_snap['name']

    def _get_coprhd_volume_name(self, vol, verbose=False):
        tagname = self.OPENSTACK_TAG + ":id:" + vol.id
        rslt = coprhd_utils.search_by_tag(
            coprhd_vol.Volume.URI_SEARCH_VOLUMES_BY_TAG.format(tagname),
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port)

        # if the result is empty, then search with the tagname
        # as "OpenStack:obj_id"
        # as snapshots will be having the obj_id instead of just id.
        if len(rslt) == 0:
            tagname = self.OPENSTACK_TAG + ":obj_id:" + vol.id
            rslt = coprhd_utils.search_by_tag(
                coprhd_vol.Volume.URI_SEARCH_VOLUMES_BY_TAG.format(tagname),
                self.configuration.coprhd_hostname,
                self.configuration.coprhd_port)

        if len(rslt) > 0:
            rslt_vol = self.volume_obj.show_by_uri(rslt[0])

            if verbose is True:
                return {'volume_name': rslt_vol['name'], 'volume_uri': rslt[0]}
            else:
                return rslt_vol['name']
        else:
            raise coprhd_utils.CoprHdError(
                coprhd_utils.CoprHdError.NOT_FOUND_ERR,
                (_("Volume %s not found") % vol['display_name']))

    def _get_resource_name(self, resource,
                           max_name_cap=MAX_DEFAULT_NAME_LENGTH,
                           truncate_name=False):
        # 36 refers to the length of UUID and +1 for '-'
        permitted_name_length = max_name_cap - (36 + 1)
        name = resource.display_name
        if not name:
            name = resource.name

        '''
        for scaleio, truncate_name will be true. We make sure the
        total name is less than or equal to 31 characters.
        _id_to_base64 will return a 24 character name'''
        if truncate_name:
            name = self._id_to_base64(resource.id)
            return name

        elif len(name) > permitted_name_length:
            '''
            The maximum length of resource name in CoprHD is 128. Hence we use
            only first 91 characters of the resource name'''
            return name[0:permitted_name_length] + "-" + resource.id

        else:
            return name + "-" + resource.id

    def _get_vpool(self, volume):
        vpool = {}
        ctxt = context.get_admin_context()
        type_id = volume.volume_type_id
        if type_id is not None:
            volume_type = volume_types.get_volume_type(ctxt, type_id)
            specs = volume_type.get('extra_specs')
            for key, value in specs.items():
                vpool[key] = value

        return vpool

    def _id_to_base64(self, id):
        # Base64 encode the id to get a volume name less than 32 characters due
        # to ScaleIO limitation.
        name = six.text_type(id).replace("-", "")
        try:
            name = base64.b16decode(name.upper())
        except (TypeError, binascii.Error):
            pass
        encoded_name = name
        if isinstance(encoded_name, six.text_type):
            encoded_name = encoded_name.encode('utf-8')
        encoded_name = base64.b64encode(encoded_name)
        if six.PY3:
            encoded_name = encoded_name.decode('ascii')
        LOG.debug("Converted id %(id)s to scaleio name %(name)s.",
                  {'id': id, 'name': encoded_name})
        return encoded_name

    def _raise_or_log_exception(self, err_code, coprhd_err_msg, log_err_msg):

        if err_code == coprhd_utils.CoprHdError.SOS_FAILURE_ERR:
            raise coprhd_utils.CoprHdError(
                coprhd_utils.CoprHdError.SOS_FAILURE_ERR,
                coprhd_err_msg)
        else:
            with excutils.save_and_reraise_exception():
                LOG.exception(log_err_msg)

    @retry_wrapper
    def _find_exportgroup(self, initiator_ports):
        """Find export group with initiator ports same as given initiators."""
        foundgroupname = None
        grouplist = self.exportgroup_obj.exportgroup_list(
            self.configuration.coprhd_project,
            self.configuration.coprhd_tenant)
        for groupid in grouplist:
            groupdetails = self.exportgroup_obj.exportgroup_show(
                groupid,
                self.configuration.coprhd_project,
                self.configuration.coprhd_tenant)
            if groupdetails is not None:
                if groupdetails['inactive']:
                    continue
                initiators = groupdetails['initiators']
                if initiators is not None:
                    inits_eg = set()
                    for initiator in initiators:
                        inits_eg.add(initiator['initiator_port'])

                    if inits_eg <= set(initiator_ports):
                        foundgroupname = groupdetails['name']
                    if foundgroupname is not None:
                        # Check the associated varray
                        if groupdetails['varray']:
                            varray_uri = groupdetails['varray']['id']
                            varray_details = self.varray_obj.varray_show(
                                varray_uri)
                            if varray_details['name'] == (
                                    self.configuration.coprhd_varray):
                                LOG.debug(
                                    "Found exportgroup %s",
                                    foundgroupname)
                                break

                        # Not the right varray
                        foundgroupname = None

        return foundgroupname

    @retry_wrapper
    def _find_host(self, initiator_port):
        """Find the host, if exists, to which the given initiator belong."""
        foundhostname = None
        hosts = self.host_obj.list_all(self.configuration.coprhd_tenant)
        for host in hosts:
            initiators = self.host_obj.list_initiators(host['id'])
            for initiator in initiators:
                if initiator_port == initiator['name']:
                    foundhostname = host['name']
                    break

            if foundhostname is not None:
                break

        return foundhostname

    @retry_wrapper
    def get_exports_count_by_initiators(self, initiator_ports):
        """Fetches ITL map for a given list of initiator ports."""
        comma_delimited_initiator_list = ",".join(initiator_ports)
        (s, h) = coprhd_utils.service_json_request(
            self.configuration.coprhd_hostname,
            self.configuration.coprhd_port, "GET",
            URI_BLOCK_EXPORTS_FOR_INITIATORS.format(
                comma_delimited_initiator_list),
            None)

        export_itl_maps = coprhd_utils.json_decode(s)

        if export_itl_maps is None:
            return 0

        itls = export_itl_maps['itl']
        return itls.__len__()

    @retry_wrapper
    def update_volume_stats(self):
        """Retrieve stats info."""
        LOG.debug("Updating volume stats")
        self.authenticate_user()

        try:
            self.stats['consistencygroup_support'] = True
            self.stats['consistent_group_snapshot_enabled'] = True
            vols = self.volume_obj.list_volumes(
                self.configuration.coprhd_tenant +
                "/" +
                self.configuration.coprhd_project)

            vpairs = set()
            if len(vols) > 0:
                for vol in vols:
                    if vol:
                        vpair = (vol["vpool"]["id"], vol["varray"]["id"])
                        if vpair not in vpairs:
                            vpairs.add(vpair)

            if len(vpairs) > 0:
                free_gb = 0.0
                used_gb = 0.0
                for vpair in vpairs:
                    if vpair:
                        (s, h) = coprhd_utils.service_json_request(
                            self.configuration.coprhd_hostname,
                            self.configuration.coprhd_port,
                            "GET",
                            URI_VPOOL_VARRAY_CAPACITY.format(vpair[0],
                                                             vpair[1]),
                            body=None)
                        capacity = coprhd_utils.json_decode(s)

                        free_gb += float(capacity["free_gb"])
                        used_gb += float(capacity["used_gb"])

                self.stats['free_capacity_gb'] = free_gb
                self.stats['total_capacity_gb'] = free_gb + used_gb
                self.stats['reserved_percentage'] = (
                    self.configuration.reserved_percentage)

            return self.stats

        except coprhd_utils.CoprHdError:
            with excutils.save_and_reraise_exception():
                LOG.exception("Update volume stats failed")

    @retry_wrapper
    def retype(self, ctxt, volume, new_type, diff, host):
        """changes the vpool type."""
        self.authenticate_user()
        volume_name = self._get_coprhd_volume_name(volume)
        vpool_name = new_type['extra_specs']['CoprHD:VPOOL']

        try:
            full_project_name = "%s/%s" % (
                self.configuration.coprhd_tenant,
                self.configuration.coprhd_project)

            task = self.volume_obj.update(
                full_project_name,
                volume_name,
                vpool_name)

            self.volume_obj.check_for_sync(task['task'][0], True)
            return True
        except coprhd_utils.CoprHdError as e:
            coprhd_err_msg = (_("Volume %(volume_name)s: update failed"
                                "\n%(err)s") % {'volume_name': volume_name,
                                                'err': six.text_type(e.msg)})

            log_err_msg = ("Volume : %s type update failed" %
                           volume_name)
            self._raise_or_log_exception(e.err_code, coprhd_err_msg,
                                         log_err_msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import Column
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table


def upgrade(migrate_engine):
    """Add service_uuid column to volumes."""
    meta = MetaData(bind=migrate_engine)

    Table('services', meta, autoload=True)
    volumes = Table('volumes', meta, autoload=True)
    if not hasattr(volumes.c, 'service_uuid'):
        volumes.create_column(Column('service_uuid', String(36),
                                     ForeignKey('services.uuid'),
                                     nullable=True))

    index_name = 'volumes_service_uuid_idx'
    indexes = Inspector(migrate_engine).get_indexes('volumes')
    if index_name not in (i['name'] for i in indexes):
        volumes = Table('volumes', meta, autoload=True)
        Index(index_name, volumes.c.service_uuid, volumes.c.deleted).create()



# Copyright (c) 2016 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest

from cinder.tests.unit.volume.drivers.dell_emc.unity \
    import fake_exception as ex
from cinder.tests.unit.volume.drivers.dell_emc.unity import test_adapter
from cinder.volume import configuration as conf
from cinder.volume.drivers.dell_emc.unity import driver


########################
#
#   Start of Mocks
#
########################

class MockAdapter(object):
    def do_setup(self, driver_object, configuration):
        raise ex.AdapterSetupError()

    @staticmethod
    def create_volume(volume):
        return volume

    @staticmethod
    def create_volume_from_snapshot(volume, snapshot):
        return volume

    @staticmethod
    def create_cloned_volume(volume, src_vref):
        return volume

    @staticmethod
    def extend_volume(volume, new_size):
        volume.size = new_size

    @staticmethod
    def delete_volume(volume):
        volume.exists = False

    @staticmethod
    def create_snapshot(snapshot):
        snapshot.exists = True
        return snapshot

    @staticmethod
    def delete_snapshot(snapshot):
        snapshot.exists = False

    @staticmethod
    def initialize_connection(volume, connector):
        return {'volume': volume, 'connector': connector}

    @staticmethod
    def terminate_connection(volume, connector):
        return {'volume': volume, 'connector': connector}

    @staticmethod
    def update_volume_stats():
        return {'stats': 123}

    @staticmethod
    def manage_existing(volume, existing_ref):
        volume.managed = True
        return volume

    @staticmethod
    def manage_existing_get_size(volume, existing_ref):
        volume.managed = True
        volume.size = 7
        return volume

    @staticmethod
    def get_pool_name(volume):
        return 'pool_0'

    @staticmethod
    def initialize_connection_snapshot(snapshot, connector):
        return {'snapshot': snapshot, 'connector': connector}

    @staticmethod
    def terminate_connection_snapshot(snapshot, connector):
        return {'snapshot': snapshot, 'connector': connector}

    @staticmethod
    def restore_snapshot(volume, snapshot):
        return True

########################
#
#   Start of Tests
#
########################


class UnityDriverTest(unittest.TestCase):
    @staticmethod
    def get_volume():
        return test_adapter.MockOSResource(provider_location='id^lun_43',
                                           id='id_43')

    @classmethod
    def get_snapshot(cls):
        return test_adapter.MockOSResource(volume=cls.get_volume())

    @staticmethod
    def get_context():
        return None

    @staticmethod
    def get_connector():
        return {'host': 'host1'}

    def setUp(self):
        self.config = conf.Configuration(None)
        self.driver = driver.UnityDriver(configuration=self.config)
        self.driver.adapter = MockAdapter()

    def test_default_initialize(self):
        config = conf.Configuration(None)
        iscsi_driver = driver.UnityDriver(configuration=config)
        self.assertListEqual([], config.unity_storage_pool_names)
        self.assertListEqual([], config.unity_io_ports)
        self.assertTrue(config.san_thin_provision)
        self.assertEqual('', config.san_ip)
        self.assertEqual('admin', config.san_login)
        self.assertEqual('', config.san_password)
        self.assertEqual('', config.san_private_key)
        self.assertEqual('', config.san_clustername)
        self.assertEqual(22, config.san_ssh_port)
        self.assertEqual(False, config.san_is_local)
        self.assertEqual(30, config.ssh_conn_timeout)
        self.assertEqual(1, config.ssh_min_pool_conn)
        self.assertEqual(5, config.ssh_max_pool_conn)
        self.assertEqual('iSCSI', iscsi_driver.protocol)

    def test_fc_initialize(self):
        config = conf.Configuration(None)
        config.storage_protocol = 'fc'
        fc_driver = driver.UnityDriver(configuration=config)
        self.assertEqual('FC', fc_driver.protocol)

    def test_do_setup(self):
        def f():
            self.driver.do_setup(None)

        self.assertRaises(ex.AdapterSetupError, f)

    def test_create_volume(self):
        volume = self.get_volume()
        self.assertEqual(volume, self.driver.create_volume(volume))

    def test_create_volume_from_snapshot(self):
        volume = self.get_volume()
        snap = self.get_snapshot()
        self.assertEqual(
            volume, self.driver.create_volume_from_snapshot(volume, snap))

    def test_create_cloned_volume(self):
        volume = self.get_volume()
        self.assertEqual(
            volume, self.driver.create_cloned_volume(volume, None))

    def test_extend_volume(self):
        volume = self.get_volume()
        self.driver.extend_volume(volume, 6)
        self.assertEqual(6, volume.size)

    def test_delete_volume(self):
        volume = self.get_volume()
        self.driver.delete_volume(volume)
        self.assertFalse(volume.exists)

    def test_create_snapshot(self):
        snapshot = self.get_snapshot()
        self.driver.create_snapshot(snapshot)
        self.assertTrue(snapshot.exists)

    def test_delete_snapshot(self):
        snapshot = self.get_snapshot()
        self.driver.delete_snapshot(snapshot)
        self.assertFalse(snapshot.exists)

    def test_ensure_export(self):
        self.assertIsNone(self.driver.ensure_export(
            self.get_context(), self.get_volume()))

    def test_create_export(self):
        self.assertIsNone(self.driver.create_export(
            self.get_context(), self.get_volume(), self.get_connector()))

    def test_remove_export(self):
        self.assertIsNone(self.driver.remove_export(
            self.get_context(), self.get_volume()))

    def test_check_for_export(self):
        self.assertIsNone(self.driver.check_for_export(
            self.get_context(), self.get_volume()))

    def test_initialize_connection(self):
        volume = self.get_volume()
        connector = self.get_connector()
        conn_info = self.driver.initialize_connection(volume, connector)
        self.assertEqual(volume, conn_info['volume'])
        self.assertEqual(connector, conn_info['connector'])

    def test_terminate_connection(self):
        volume = self.get_volume()
        connector = self.get_connector()
        conn_info = self.driver.terminate_connection(volume, connector)
        self.assertEqual(volume, conn_info['volume'])
        self.assertEqual(connector, conn_info['connector'])

    def test_update_volume_stats(self):
        stats = self.driver.get_volume_stats(True)
        self.assertEqual(123, stats['stats'])
        self.assertEqual(self.driver.VERSION, stats['driver_version'])
        self.assertEqual(self.driver.VENDOR, stats['vendor_name'])

    def test_manage_existing(self):
        volume = self.driver.manage_existing(self.get_volume(), None)
        self.assertTrue(volume.managed)

    def test_manage_existing_get_size(self):
        volume = self.driver.manage_existing_get_size(self.get_volume(), None)
        self.assertTrue(volume.managed)
        self.assertEqual(7, volume.size)

    def test_get_pool(self):
        self.assertEqual('pool_0', self.driver.get_pool(self.get_volume()))

    def test_unmanage(self):
        ret = self.driver.unmanage(None)
        self.assertIsNone(ret)

    def test_backup_use_temp_snapshot(self):
        self.assertTrue(self.driver.backup_use_temp_snapshot())

    def test_initialize_connection_snapshot(self):
        snapshot = self.get_snapshot()
        conn_info = self.driver.initialize_connection_snapshot(
            snapshot, self.get_connector())
        self.assertEqual(snapshot, conn_info['snapshot'])

    def test_terminate_connection_snapshot(self):
        snapshot = self.get_snapshot()
        conn_info = self.driver.terminate_connection_snapshot(
            snapshot, self.get_connector())
        self.assertEqual(snapshot, conn_info['snapshot'])

    def test_restore_snapshot(self):
        snapshot = self.get_snapshot()
        volume = self.get_volume()
        r = self.driver.revert_to_snapshot(None, volume, snapshot)
        self.assertTrue(r)
